#! /usr/bin/env python3.9

"""
This module helps you find leased IP addresses of a country and make sure that when
users are trying to connect to a website in that country, they will be connected directly.

GitHub: https://github.com/Mahyar24/V2Conf
Mahyar@Mahyar24.com, Sun Nov 13 2022
"""
import logging
from typing import Union

import requests


def get_ips(country_code: str, logger: logging.Logger) -> list[str]:
    """
    Get leased IP addresses of a country.
    Based on country codes (ISO 3166-1 alpha-2).
    """
    link = (
        f"https://stat.ripe.net/data/country-resource-list/data.json"
        f"?resource={country_code}&v4_format=prefix"
    )
    for _ in range(5):
        try:
            resp = requests.get(link, timeout=20)
            resp.raise_for_status()
        except requests.exceptions.RequestException:
            pass
        else:
            try:
                return resp.json()["data"]["resources"]["ipv4"]
            except KeyError:
                logger.error(f"{link!r} doesn't respond on planned schema")
                return []

    logger.error(f"Failed to connect to {link!r}")
    return []


def make_ip_rule(
    country_code: str, freedom_tag: str, logger: logging.Logger
) -> dict[str, Union[str, list[str]]]:
    """
    Make a rule to exclude IP addresses of a country.
    """
    return {
        "type": "field",
        "ip": get_ips(country_code, logger),
        "outboundTag": freedom_tag,
    }



#! /usr/bin/env python3.9

"""
This module is used for reading and writing configuration files.

GitHub: https://github.com/Mahyar24/V2Conf
Mahyar@Mahyar24.com, Sun Nov 13 2022
"""


import argparse
import json
import logging
import random
import secrets
import socket
from contextlib import closing
from pathlib import Path
from typing import Optional, Union

from .exclude import make_ip_rule


def find_n_unused_port(num: int) -> list[int]:
    """
    Find n unique unused port and return them as a list.
    """
    ports: set[int] = set()
    while len(ports) < num:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            sock.bind(("", 0))
            port = sock.getsockname()[1]
            ports.add(port)

    return list(ports)


def make_http_inbounds(outbound_tags: list[str], logger: logging.Logger) -> list[dict]:
    """
    For every outbound, create a http inbound for examining the performance.
    """
    http_inbounds = []

    ports = find_n_unused_port(len(outbound_tags))

    for port, outbound_tag in zip(ports, outbound_tags):
        http_inbounds.append(
            {
                "port": port,
                "listen": "127.0.0.1",
                "protocol": "http",
                "accounts": [
                    {
                        "user": secrets.token_urlsafe(16),
                        "pass": secrets.token_urlsafe(16),
                    }
                ],
                "tag": f"inbound-http-test-{outbound_tag}",
            }
        )
        logger.info(f"HTTP proxy created '127.0.0.1:{port}' for {outbound_tag!r}")

    return http_inbounds


def read_inbounds(path: Path) -> dict[str, dict]:
    """
    Reading inbounds from the path.
    """
    inbounds_path = path / "inbounds"

    if not inbounds_path.is_dir():
        raise ValueError("Inbounds directory not found!")

    inbounds = {}
    tags = []
    for config in inbounds_path.glob("*.json"):
        with open(config, "r", encoding="utf-8") as file:
            data = json.load(file)
        try:
            inbounds[data["tag"]] = data
        except KeyError as err:
            raise KeyError(f"Tag not found in {config.name!r}") from err
        else:
            if data["tag"].startswith("inbound-http-test-"):
                raise ValueError(
                    f"'inbound-http-test-*' is reserved for internal use. "
                    f"Change the tag in {config.name!r}"
                )

            if (tag := data["tag"]) in tags:
                raise ValueError(
                    f"Duplicated tags; {tag!r} already exist! Change the tag in {config.name!r}"
                )
            tags.append(tag)

    if len(inbounds) < 1:
        raise ValueError("No inbound config found!")

    return inbounds


def read_outbounds(path: Path) -> dict[str, dict]:
    """
    Reading outbounds from the path. at least one freedom outbound is required.
    """
    outbounds_path = path / "outbounds"

    if not outbounds_path.is_dir():
        raise ValueError("Outbounds directory not found!")

    outbounds = {}
    tags = []
    check_freedom = False
    for config in outbounds_path.glob("*.json"):
        with open(config, "r", encoding="utf-8") as file:
            data = json.load(file)
        try:
            outbounds[data["tag"]] = data
        except KeyError as err:
            raise KeyError(f"Tag not found in {config.name!r}") from err
        else:
            if (tag := data["tag"]) in tags:
                raise ValueError(
                    f"Duplicated tags; {tag!r} already exist! Change the tag in {config.name!r}"
                )
            tags.append(tag)

        if data["protocol"] == "freedom":
            check_freedom = True

    if not check_freedom:
        raise ValueError("Freedom outbound not found!")

    return outbounds


def read_rules(path: Path) -> dict[str, dict]:
    """
    Reading rules from the path.
    """
    rules_path = path / "rules"

    if not rules_path.is_dir():
        raise ValueError("Rules directory not found!")

    rules = {}
    tags = []
    for config in rules_path.glob("*.json"):
        with open(config, "r", encoding="utf-8") as file:
            data = json.load(file)
        try:
            rules[data["tag"]] = data
        except KeyError as err:
            raise KeyError(f"Tag not found in {config.name!r}") from err
        else:
            if (tag := data["tag"]) in tags:
                raise ValueError(
                    f"Duplicated tags; {tag!r} already exist! Change the tag in {config.name!r}"
                )
            tags.append(tag)

    return rules


def make_rules(
    path: Path,
    logger: logging.Logger,
    *,
    geoip: bool,
    county_code: str,
    http_inbounds: list[dict],
    inbound_tags: list[str],
    freedom_outbound_tag: str,
    outbound: str,
) -> list[dict]:
    """
    Make rules for the configuration file.
    """

    # Rules for excluding domestic IPs.
    if geoip:
        ip_rule: dict[str, Union[str, list[str]]] = {
            "type": "field",
            "outboundTag": freedom_outbound_tag,
            "ip": [f"geoip:{county_code.lower()}"],
        }
        logger.info(f"'geoip:{county_code.lower()}' going to {freedom_outbound_tag!r}")
    else:
        ip_rule = make_ip_rule(county_code, freedom_outbound_tag, logger)
        logger.info(f"Excluding {len(ip_rule['ip'])} IPs for {county_code!r}")

    # Rules from rules directory
    rules = list(read_rules(path).values())
    logger.info(f"Read {len(rules)} rules")

    # If downloading IPs from the internet failed, we must not
    # create a rule with empty list, otherwise v2ray will crash.
    if ip_rule["ip"]:
        rules += [ip_rule]

    # Rules for routing http proxies to their corresponding outbounds.
    for http_inbound in http_inbounds:
        rules.append(
            {
                "inboundTag": [http_inbound["tag"]],
                "outboundTag": http_inbound["tag"].removeprefix("inbound-http-test-"),
                "type": "field",
            }
        )

    logger.info(f"Using {outbound!r} as main outbound")
    rules.append(
        {
            "inboundTag": inbound_tags,
            "outboundTag": outbound,
            "type": "field",
        }
    )

    return rules


def make_conf(
    args: argparse.Namespace,
    logger: logging.Logger,
    outbound_tag: Optional[str] = None,
) -> dict:
    """
    Make a new configuration file.
    """
    inbounds = read_inbounds(args.path_conf_dir)
    logger.info(f"Read {len(inbounds)} inbounds")

    outbounds = read_outbounds(args.path_conf_dir)

    # We should only make http inbounds for the outbounds that are not freedom or blackhole.
    vpn_outbounds = [
        k for k, v in outbounds.items() if v["protocol"] not in ("freedom", "blackhole")
    ]
    if len(vpn_outbounds) < 2:
        raise ValueError(
            "At least two non-freedom/non-blackhole outbounds are required!"
        )

    logger.info(f"Read {len(outbounds)} outbounds, {len(vpn_outbounds)} vpn outbounds")

    # find tag of freedom outbound, using when making first naive config file.
    freedom = [k for k, v in outbounds.items() if v["protocol"] == "freedom"][0]

    http_inbounds = make_http_inbounds(vpn_outbounds, logger)

    # If nothing is specified for `outbound_tag`, route all traffic to
    # a random vpn outbound. (cold start), Otherwise, route all traffic
    # to the specified (best) outbound.
    if outbound_tag is None:
        logger.info("Cold start, routing all traffic to a random vpn outbound")
        outbound_tag = random.choice(vpn_outbounds)

    rules = make_rules(
        args.path_conf_dir,
        logger,
        geoip=not args.no_geoip,
        county_code=args.country_code,
        http_inbounds=http_inbounds,
        inbound_tags=list(inbounds.keys()),
        freedom_outbound_tag=freedom,
        outbound=outbound_tag,
    )

    return {
        "log": {"loglevel": args.log_level},
        "inbounds": list(inbounds.values()) + http_inbounds,
        "outbounds": list(outbounds.values()),
        "routing": {"domainStrategy": "IPIfNonMatch", "rules": rules},
    }


def write_conf(path: Path, conf: dict) -> None:
    """
    Rewrite the configuration file.
    """
    with open(path, "w", encoding="utf-8") as file:
        json.dump(conf, file, indent=4)



#! /usr/bin/env python3.9

"""
This package helps you to create the configuration file for V2Ray automatically
and find the best outbounds and change the routing rules periodically based on
outbounds performances.
Dependencies:
    - requests
    - aiohttp
    - [jdatetime] (optional) (if you want to configure the Jalali date-time for logging)

Check out this Gist for detecting abnormal user traffic:
https://gist.github.com/Mahyar24/d712a30a35576e5b8584c562e15e550c

Compatible with Python 3.9+
GitHub: https://github.com/Mahyar24/V2Conf
Mahyar@Mahyar24.com, Sun Nov 13 2022
"""

import argparse
import asyncio
import logging
import os
import re
import subprocess
import textwrap
import time
from pathlib import Path
from time import struct_time
from zoneinfo import ZoneInfo

from .configs import make_conf, write_conf
from .health import rank_outbounds

__version__ = "0.0.8"
__author__ = "Mahyar Mahdavi"
__email__ = "Mahyar@Mahyar24.com"
__license__ = "GPLv3"
__url__ = "https://GitHub.com/Mahyar24/V2Conf"
__pypi__ = "https://PyPI.org/project/V2Conf"


def check_requirements() -> None:
    """
    Assert if user is running the script with root privileges.
    """
    assert os.getuid() == 0, "You must have super user permissions to run this program."


def tehran_time(*_, **__) -> time.struct_time:
    """
    Struct_time for Jalali date-time.
    """

    try:
        import jdatetime  # type: ignore
    except ImportError as err:
        raise ImportError(
            "Please install `jdatetime` module to use Jalali date-time for logging"
        ) from err

    tehran_dt = jdatetime.datetime.now(tz=ZoneInfo("Asia/Tehran"))
    return struct_time(
        (
            tehran_dt.year,
            tehran_dt.month,
            tehran_dt.day,
            tehran_dt.hour,
            tehran_dt.minute,
            tehran_dt.second,
            tehran_dt.weekday(),
            tehran_dt.yday(),
            1,
        )
    )


def make_logger(args: argparse.Namespace) -> logging.Logger:
    """
    Make a logger and return it.
    """
    logging.raiseExceptions = False
    if args.jalali:
        logging.Formatter.converter = tehran_time

    logger = logging.getLogger(os.path.basename(__file__))
    logger.setLevel(logging.INFO)

    handler_formatter = logging.Formatter(
        "%(asctime)s:%(levelname)s:%(message)s.", datefmt="%Y-%m-%d %H:%M:%S"
    )

    stdout_handler = logging.StreamHandler()
    stdout_handler.setFormatter(handler_formatter)
    logger.addHandler(stdout_handler)

    if not args.quiet:
        file_handler = logging.FileHandler(args.log_file)
        file_handler.setFormatter(handler_formatter)
        logger.addHandler(file_handler)

    return logger


def validate_url(text: str) -> bool:
    """
    Check if the text is a valid http(s) URL.
    """
    pattern = re.compile(
        r"^(http)s?://"
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
        r"localhost|"
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
        r"(?::\d+)?"
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    return bool(re.match(pattern, text))


def checking_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    """
    Check the passed arguments and return the parsed arguments.
    """
    args = parser.parse_args()
    if not validate_url(args.website):
        parser.error("The website you entered is not a valid URL.")
    if not args.path_conf_dir.is_dir():
        parser.error(
            f"The directory you entered {args.path_conf_dir!r} is not a valid directory."
        )
    return args


def parsing_args() -> argparse.Namespace:
    """
    Parsing the passed arguments, read help (-h, --help) for further information.
    """
    parser = argparse.ArgumentParser(
        epilog=textwrap.dedent(
            """
            Written by: Mahyar Mahdavi <Mahyar@Mahyar24.com>.
            License: GNU GPLv3.
            Source Code: <https://github.com/mahyar24/V2Conf>.
            Reporting Bugs and PRs are welcomed. :)
            """
        )
    )
    group = parser.add_mutually_exclusive_group()

    parser.add_argument(
        "path_conf_dir",
        nargs="?",
        default=Path.cwd(),
        type=Path,
        help="Select configuration directory, default is $PWD.",
    )

    parser.add_argument(
        "-c",
        "--config-file",
        default=Path("/usr/local/etc/v2ray/config.json"),
        type=Path,
        help="Select configuration file, default is '/usr/local/etc/v2ray/config.json'.",
    )

    parser.add_argument(
        "--country-code",
        help="Exclude a country from the list of IPs to be routed; "
        "default is 'IR'. (ISO 3166-1 alpha-2)",
        type=lambda x: x.upper(),
        default="IR",
    )

    parser.add_argument(
        "--no-geoip",
        help="Instead of using V2Ray GeoIP database, "
        "downloading IPs from 'ripe.net' (more recent IPs but may slow V2Ray)",
        action="store_true",
    )

    parser.add_argument(
        "-t",
        "--timeout",
        help="Set the timeout for checking the health of proxies, default is 15 seconds.",
        type=int,
        default=15,
    )

    parser.add_argument(
        "-w",
        "--website",
        help="Set the website to be used for checking the health of proxies, "
        "default is 'https://facebook.com'",
        default="https://facebook.com",
    )

    parser.add_argument(
        "-n",
        "--num-of-tries",
        help="Set the number of tries for checking the health of proxies, default is 10.",
        type=int,
        default=10,
    )

    parser.add_argument(
        "-s",
        "--sleep-time",
        help="Set the sleep time between each checkup, default is 1,800s. (in seconds)",
        type=int,
        default=1_800,
    )

    parser.add_argument(
        "-l",
        "--log-level",
        help="Set the V2Ray log level, default is 'warning'.",
        choices=["debug", "info", "warning", "error", "none"],
        default="warning",
    )

    group.add_argument(
        "-q",
        "--quiet",
        help="No log file (V2Conf). (printing to stdout anyway)",
        action="store_true",
    )

    group.add_argument(
        "--log-file",
        help="Path for V2Conf log file. default is '$PWD/V2Conf.log'",
        default=Path.cwd() / "V2Conf.log",
        type=Path,
    )

    parser.add_argument(
        "--jalali", help="Use Jalali datetime for V2Conf logging", action="store_true"
    )

    parser.add_argument(
        "-v",
        "--version",
        help="Show version and exit.",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    return checking_args(parser)


def restart_v2ray(logger: logging.Logger) -> None:
    """
    Restart V2Ray service. (systemctl)
    When `check=True` is passed to `run`, `CalledProcessError` is raised if the
    exit code was non-zero.
    """
    subprocess.run(["systemctl", "restart", "v2ray"], check=True)
    logger.warning("V2Ray is restarted")


def main() -> None:
    """
    Main function.
    """
    check_requirements()

    args = parsing_args()
    logger = make_logger(args)

    # At the first run, we will make a naive configuration file.
    # and all inbounds will route to a randomly selected outbound.
    conf = make_conf(args, logger)
    write_conf(args.config_file, conf)
    logger.info("Naive configuration file is written")

    restart_v2ray(logger)

    time.sleep(30)

    while True:
        # Ranking outbound performances.
        ranked_outbounds = asyncio.run(
            rank_outbounds(
                args,
                logger,
                http_inbounds=[
                    inbound
                    for inbound in conf["inbounds"]
                    if inbound["tag"].startswith("inbound-http-test-")
                ],
            )
        )

        # Make the new configuration file and set all inbounds to the best inbound.
        conf = make_conf(args, logger, ranked_outbounds[0])
        # Write the new configuration file.
        write_conf(args.config_file, conf)
        logger.info("Configuration file is written")
        # Restarting to apply the new configuration file.
        restart_v2ray(logger)
        # Sleeping until the next checkup.
        logger.info(f"Sleeping for {args.sleep_time:,} seconds")
        time.sleep(args.sleep_time)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['v2conf']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0', 'jdatetime>=4.1.0,<5.0.0', 'requests>=2.28.1,<3.0.0']

entry_points = \
{'console_scripts': ['v2conf = v2conf.__main__:main']}

setup_kwargs = {
    'name': 'v2conf',
    'version': '0.0.8',
    'description': 'V2Conf helps you build V2Ray Config file automatically and evaluate and change config rules based on outbounds performances.',
    'long_description': 'None',
    'author': 'Mahyar Mahdavi',
    'author_email': 'Mahyar@Mahyar24.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Mahyar24/V2Conf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#! /usr/bin/env python3.9

"""
In this module, we are going to check the health and performance of
all the outbounds and rank them based on their health and performance.

GitHub: https://github.com/Mahyar24/V2Conf
Mahyar@Mahyar24.com, Sun Nov 13 2022
"""

import argparse
import asyncio
import logging
import math
import statistics
import time
from collections import ChainMap

import aiohttp


async def check_health(
    session: aiohttp.ClientSession,
    outbound_name: str,
    username: str,
    password: str,
    proxy: str,
    test_site: str,
    *,
    num_of_tries: int,
    timeout: int,
) -> dict[str, list[float]]:
    """
    Check the health of a proxy by pinging it and checking its latency.
    If an error or timeout happens, it will be recorded as a NaN,
    otherwise we will record the latency.
    """
    result: list[float] = []

    proxy_auth = aiohttp.BasicAuth(username, password)
    for _ in range(num_of_tries):
        start_time = time.perf_counter()
        try:
            async with session.get(
                test_site, proxy=proxy, proxy_auth=proxy_auth, timeout=timeout
            ) as resp:
                if resp.ok:
                    # Fetch the data chunk by chuck
                    async for _ in resp.content.iter_chunked(2048):
                        pass
                    result.append(time.perf_counter() - start_time)
                else:
                    result.append(float("nan"))
        except (
            asyncio.exceptions.TimeoutError,
            aiohttp.client_exceptions.ClientError,
        ):
            result.append(float("nan"))

    return {outbound_name: result}


def calculate(results: list[float]) -> tuple[int, float]:
    """
    Calculate the health of an outbound.
    Returns a tuple of (number of errors, average latency).
    """
    number_of_nans = sum(map(math.isnan, results))
    if number_of_nans == len(results):
        return number_of_nans, float("nan")
    return number_of_nans, statistics.mean(filter(lambda x: not math.isnan(x), results))


async def rank_outbounds(
    args: argparse.Namespace, logger: logging.Logger, *, http_inbounds: list[dict]
) -> list[str]:
    """
    Rank the outbounds by their health. best to worst.
    We will sort the outbounds first by number of errors, then by average latency.
    """
    coroutines = []

    async with aiohttp.ClientSession() as session:
        for http_inbound in http_inbounds:
            coroutines.append(
                asyncio.create_task(
                    check_health(
                        session,
                        http_inbound["tag"].split("-")[-1],
                        http_inbound["accounts"][0]["user"],
                        http_inbound["accounts"][0]["pass"],
                        f"http://{http_inbound['listen']}:{http_inbound['port']}",
                        args.website,
                        num_of_tries=args.num_of_tries,
                        timeout=args.timeout,
                    )
                )
            )
        results = dict(ChainMap(*await asyncio.gather(*coroutines)))

    results = {k: calculate(v) for k, v in results.items()}
    logger.info(f"Results: {results}")

    return list(dict(sorted(results.items(), key=lambda item: item[1])).keys())
